//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Views/SampleDesigner/LayerOrientedSampleEditor.cpp
//! @brief     Implements class LayerOrientedSampleEditor
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Views/SampleDesigner/LayerOrientedSampleEditor.h"
#include "GUI/Models/LayerItem.h"
#include "GUI/Models/MultiLayerItem.h"
#include "GUI/Views/CommonWidgets/StyledToolBar.h"
#include "GUI/Views/SampleDesigner/LayerEditorUtils.h"
#include "GUI/Views/SampleDesigner/MultiLayerForm.h"
#include "GUI/Views/SampleDesigner/SampleEditorController.h"

#include <QBoxLayout>
#include <QScrollArea>

LayerOrientedSampleEditor::LayerOrientedSampleEditor(QWidget* parent)
    : QWidget(parent)
    , m_undoAction(nullptr)
    , m_redoAction(nullptr)
    , m_showInlineEditButtonsAction(nullptr)
    , m_asAngstrom(nullptr)
    , m_asRad(nullptr)
    , m_currentMultiLayerWidget(nullptr)
    , m_currentMultiLayerItem(nullptr)
{
    setAttribute(Qt::WA_StyledBackground, true);
    setProperty("stylable", true); // for stylesheet addressing

    m_scrollArea = new QScrollArea(this);
    m_scrollArea->setWidgetResizable(true);
    m_scrollArea->setWidget(createEmptyWidget());

    m_toolbar = new StyledToolBar(this);
    m_toolbar->setToolButtonStyle(Qt::ToolButtonTextBesideIcon);

    QVBoxLayout* mainLayout = new QVBoxLayout(this);
    mainLayout->setContentsMargins(0, 0, 0, 0);
    mainLayout->setSpacing(0);
    mainLayout->addWidget(m_toolbar);
    mainLayout->addWidget(m_scrollArea);

    m_showInlineEditButtonsAction = new QAction(QIcon(":/images/edit.svg"), "Edit structure", this);
    connect(m_showInlineEditButtonsAction, &QAction::toggled, this,
            &LayerOrientedSampleEditor::onShowInlineEditButtonsToggled);
    m_showInlineEditButtonsAction->setCheckable(true);
    m_showInlineEditButtonsAction->setChecked(false);
    m_toolbar->addAction(m_showInlineEditButtonsAction);

    m_asAngstrom = new QAction("A", this);
    connect(m_asAngstrom, &QAction::toggled, this, &LayerOrientedSampleEditor::onAsAngstromToggled);
    m_asAngstrom->setCheckable(true);
    m_asAngstrom->setChecked(false);
    m_toolbar->addAction(m_asAngstrom);

    m_asRad = new QAction("rad", this);
    connect(m_asRad, &QAction::toggled, this, &LayerOrientedSampleEditor::onAsRadToggled);
    m_asRad->setCheckable(true);
    m_asRad->setChecked(false);
    m_toolbar->addAction(m_asRad);
}

LayerOrientedSampleEditor::~LayerOrientedSampleEditor()
{
    qDeleteAll(m_editControllers.values());
}

void LayerOrientedSampleEditor::setCurrentSample(MultiLayerItem* multiLayerItem)
{
    if (m_currentMultiLayerItem != nullptr)
        m_editControllers[m_currentMultiLayerItem]->setMultiLayerForm(nullptr);

    m_currentMultiLayerWidget = nullptr;
    delete m_scrollArea->takeWidget();

    // use the correct toolbar buttons (from correct undo stack)
    delete m_undoAction;
    m_undoAction = nullptr;
    delete m_redoAction;
    m_redoAction = nullptr;

    m_currentMultiLayerItem = multiLayerItem;
    if (m_currentMultiLayerItem == nullptr) {
        m_scrollArea->setWidget(createEmptyWidget());
        updateActionEnabling();
        return;
    }

    if (!m_editControllers.contains(m_currentMultiLayerItem))
        m_editControllers.insert(m_currentMultiLayerItem,
                                 new SampleEditorController(m_currentMultiLayerItem));
    auto* ec = m_editControllers[m_currentMultiLayerItem];

    m_undoAction = ec->undoStack()->createUndoAction(this);
    m_redoAction = ec->undoStack()->createRedoAction(this);
    m_undoAction->setIcon(QIcon(":/images/undo-variant.svg"));
    m_redoAction->setIcon(QIcon(":/images/redo.svg"));

    m_toolbar->addAction(m_undoAction);
    m_toolbar->addAction(m_redoAction);

    createLayerColors();

    m_currentMultiLayerWidget = new MultiLayerForm(this, m_currentMultiLayerItem, ec);
    ec->setMultiLayerForm(m_currentMultiLayerWidget);
    m_currentMultiLayerWidget->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Minimum);
    m_scrollArea->setWidget(m_currentMultiLayerWidget);
    m_currentMultiLayerWidget->showInlineEditButtons(m_showInlineEditButtonsAction->isChecked());

    updateActionEnabling();
}

void LayerOrientedSampleEditor::updateActionEnabling()
{
    // #baLayerEditor implement updateActionEnabling()
}

void LayerOrientedSampleEditor::onShowInlineEditButtonsToggled(bool checked)
{
    if (m_currentMultiLayerWidget)
        m_currentMultiLayerWidget->showInlineEditButtons(checked);
}

void LayerOrientedSampleEditor::onAsAngstromToggled()
{
    if (m_currentMultiLayerWidget)
        m_currentMultiLayerWidget->useAngstrom(m_asAngstrom->isChecked());
}

void LayerOrientedSampleEditor::onAsRadToggled()
{
    if (m_currentMultiLayerWidget)
        m_currentMultiLayerWidget->useRadiant(m_asRad->isChecked());
}

void LayerOrientedSampleEditor::createLayerColors() // #baLayerEditor move to better place
{
    if (!m_currentMultiLayerItem)
        return;

    int col = 0;
    for (auto l : m_currentMultiLayerItem->layers()) {
        if (l->color().isValid())
            continue;

        l->setColor(LayerEditorUtils::predefinedLayerColors()[col]);
        col++;
        if (col == LayerEditorUtils::predefinedLayerColors().size())
            col = 0;
    }
}

QWidget* LayerOrientedSampleEditor::createEmptyWidget()
{
    QWidget* emptyWidget = new QWidget(this);
    emptyWidget->setStyleSheet("background-color: white");
    return emptyWidget;
}
