//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Views/SampleDesigner/MaterialInplaceForm.cpp
//! @brief     Implements class MaterialInplaceForm
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Views/SampleDesigner/MaterialInplaceForm.h"
#include "GUI/Models/DoubleDescriptor.h"
#include "GUI/Models/ItemWithMaterial.h"
#include "GUI/Models/MaterialItemUtils.h"
#include "GUI/Models/VectorDescriptor.h"
#include "GUI/Views/CommonWidgets/DoubleLineEdit.h"
#include "GUI/Views/MaterialEditor/MaterialEditorDialog.h"
#include "GUI/Views/SampleDesigner/LayerEditorUtils.h"
#include "GUI/Views/SampleDesigner/SampleEditorController.h"
#include "GUI/mainwindow/projectmanager.h"
#include "GUI/utils/LayoutUtils.h"

#include <QGridLayout>
#include <QLabel>
#include <QPushButton>

MaterialInplaceForm::MaterialInplaceForm(QWidget* parent, ItemWithMaterial* item,
                                         SampleEditorController* ec)
    : QWidget(parent), m_item(item), m_ec(ec)
{
    m_layout = new QGridLayout(this);
    m_layout->setContentsMargins(0, 0, 0, 0);
    createWidgets();
}

void MaterialInplaceForm::selectMaterial()
{
    const QString newMaterialIdentifier = MaterialEditorDialog::chooseMaterial(
        ProjectManager::instance()->document(), m_item->materialIdentifier());

    if (!newMaterialIdentifier.isEmpty() && newMaterialIdentifier != m_item->materialIdentifier()) {
        GUI::Utils::Layout::clearLayout(m_layout, true);
        m_ec->selectMaterial(m_item, newMaterialIdentifier);
        createWidgets();
    }
}

void MaterialInplaceForm::createWidgets()
{
    DoubleDescriptors values;

    auto* materialItem = GUI::Model::MaterialItemUtils::findMaterial(m_item->materialIdentifier());
    if (materialItem->hasRefractiveIndex())
        values << materialItem->delta() << materialItem->beta();
    else
        values << materialItem->sldRe() << materialItem->sldIm();

    int col = 0;
    for (auto d : values) {
        auto* editor = new DoubleLineEdit(this, d);
        auto* label = new QLabel(d.label, this);
        label->setBuddy(editor);

        QObject::connect(editor, &DoubleLineEdit::baseValueChanged,
                         [=](double newValue) { m_ec->setMaterialValue(m_item, newValue, d); });

        m_layout->addWidget(label, 0, col);
        m_layout->addWidget(editor, 1, col++);
    }

    LayerEditorUtils::addVectorToGrid(m_layout, col, materialItem->magnetizationVector(), m_ec,
                                      true, false);

    QPushButton* btn = new QPushButton("...", this);
    btn->setToolTip("Select material");
    m_layout->addWidget(btn, 1, m_layout->columnCount());
    connect(btn, &QPushButton::clicked, this, &MaterialInplaceForm::selectMaterial);

    m_layout->addItem(new QSpacerItem(0, 0, QSizePolicy::Expanding), 0, m_layout->columnCount());
}
