//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Views/SampleDesigner/WidgetMoverButton.cpp
//! @brief     Implements class WidgetMoverButton
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Views/SampleDesigner/WidgetMoverButton.h"
#include <QLayout>
#include <QMouseEvent>
#include <QPropertyAnimation>

WidgetMoverButton::WidgetMoverButton(QWidget* parent, QWidget* widgetToMove, int ignoreOnTop)
    : QToolButton(parent), m_widgetToMove(widgetToMove), m_ignoreOnTop(ignoreOnTop)
{
    setIcon(QIcon(":/images/toolbar24dark_hambar.svg"));
}

void WidgetMoverButton::mousePressEvent(QMouseEvent* event)
{
    m_mouseDownY = event->globalY();
}

void WidgetMoverButton::mouseReleaseEvent(QMouseEvent* event)
{
    m_started = false;
    if (m_layoutToDeactivate != nullptr) {
        m_layoutToDeactivate->setEnabled(true);
        m_layoutToDeactivate->update();
    }
    emit finishedMoving(m_widgetToMove, m_dropAboveThisWidget);
}

void WidgetMoverButton::mouseMoveEvent(QMouseEvent* event)
{
    int dy = event->globalY() - m_mouseDownY;
    if (abs(dy) > 5 && !m_started) {
        m_started = true;

        m_layoutToDeactivate = m_widgetToMove->parentWidget()->layout();

        m_yOfFirstRelevantWidget = m_layoutToDeactivate->itemAt(m_ignoreOnTop)->geometry().top();
        m_layoutToDeactivate->setEnabled(false);
        m_originalWidgetY = m_widgetToMove->y();
        m_widgetToMove->raise();
        emit startingToMove();
        m_widgetToMove->move(m_widgetToMove->x() + 5, m_widgetToMove->y());
        m_widgetToMove->resize(m_widgetToMove->width() - 10, m_widgetToMove->height());
    }

    if (m_started) {

        // -- move the dragged widget to the new position
        m_widgetToMove->move(m_widgetToMove->x(), m_originalWidgetY + dy);

        // -- move other widgets to make room
        m_dropAboveThisWidget = nullptr;
        int yTopOfLayoutItem = m_yOfFirstRelevantWidget;

        for (int i = m_ignoreOnTop; i < m_layoutToDeactivate->count(); ++i) {
            auto* layoutItem = m_layoutToDeactivate->itemAt(i);
            if (layoutItem->widget() == m_widgetToMove)
                continue;

            const bool movedWidgetIsAboveThisLayoutItem =
                m_widgetToMove->y() < yTopOfLayoutItem + layoutItem->geometry().height() / 2;
            if (movedWidgetIsAboveThisLayoutItem && layoutItem->widget() != nullptr
                && m_dropAboveThisWidget == nullptr)
                m_dropAboveThisWidget = layoutItem->widget();

            QRect r = layoutItem->geometry();
            if (movedWidgetIsAboveThisLayoutItem)
                r.moveTop(yTopOfLayoutItem + m_widgetToMove->height()
                          + m_layoutToDeactivate->spacing());
            else
                r.moveTop(yTopOfLayoutItem);

            if (r != layoutItem->geometry()) {
                QWidget* w = layoutItem->widget();
                if (w == nullptr) {
                    layoutItem->setGeometry(r);
                } else {
                    QPropertyAnimation* animation = new QPropertyAnimation(w, "geometry");
                    animation->setDuration(100);
                    animation->setEasingCurve(QEasingCurve::OutQuad);
                    animation->setStartValue(w->geometry());
                    animation->setEndValue(r);

                    animation->start();
                }
            }

            yTopOfLayoutItem += r.height() + m_layoutToDeactivate->spacing();
        }
    }
}
